from enum import Enum
from fractions import Fraction


class TypeDuration(str, Enum):
    """
    Type-duration of a note/rest (i.e. whole, quarter, 16th).
    It encodes what the musical symbol looks like, not how much time it
    actually takes up. (e.g. a triplet will be an eighth type; a grace
    note has type but no duration)

    Based on the MusicXML <type> element.
    """
    
    thousand_twenty_fourth = "1024th"
    five_hundred_twelfth = "512th"
    two_hundred_fifty_sixth = "256th"
    hundred_twenty_eighth = "128th"
    sixty_fourth = "64th"
    thirty_second = "32nd"
    sixteenth = "16th"
    eighth = "eighth"
    quarter = "quarter"
    half = "half"
    whole = "whole"
    breve = "breve"
    long = "long"
    maxima = "maxima"

    def to_quarter_multiple(self) -> Fraction:
        """Returns the number of beats (quarter notes) that this duration type
        takes up in total."""
        _LOOKUP = {
            "1024th":  Fraction(1, 256),
            "512th":   Fraction(1, 128),
            "256th":   Fraction(1, 64),
            "128th":   Fraction(1, 32),
            "64th":    Fraction(1, 16),
            "32nd":    Fraction(1, 8),
            "16th":    Fraction(1, 4),
            "eighth":  Fraction(1, 2),
            "quarter": Fraction(1, 1), # one
            "half":    Fraction(2, 1),
            "whole":   Fraction(4, 1),
            "breve":   Fraction(8, 1),
            "long":    Fraction(16, 1),
            "maxima":  Fraction(32, 1),
        }
        beats = _LOOKUP.get(self.value, None)
        if beats is None:
            raise Exception("Unknown beat count for type " + self)
        return beats



from smashcima.geometry import Vector2
from smashcima.scene import Glyph, Sprite
import matplotlib.pyplot as plt
from typing import List
import cv2


def _plot_glyph_on_axis(ax: plt.Axes, glyph: Glyph, glyph_index: int):
    ax.set_title(str(glyph_index))

    for sprite, transform in Sprite.traverse_sprites(glyph.space):
        # upper left image corner
        a = transform.apply_to(Vector2(0, 0))

        # lower right image corner
        b = transform.apply_to(Vector2(sprite.pixel_width, sprite.pixel_height))

        # NOTE: this rendering does not support rotations!
        ax.imshow(
            cv2.cvtColor(sprite.bitmap, cv2.COLOR_BGRA2RGBA),
            extent=(a.x, b.x, b.y, a.y) # (left, right, bottom, top)
        )

        ax.plot([0.0], [0.0], "rx", alpha=0.8)


def display_glyphs(rows: int, columns: int, glyphs: List[Glyph], start_from=0):
    fig, axs = plt.subplots(rows, columns, sharex=True, sharey=True)
    fig.suptitle(glyphs[0].label + f" ({len(glyphs)} total)")
    fig.tight_layout()
    
    for r in range(rows):
        for c in range(columns):
            i = start_from + r * columns + c
            if i >= len(glyphs):
                continue
            _plot_glyph_on_axis(axs[r, c], glyphs[i], i)


from smashcima.scene.LineGlyph import LineGlyph
from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from smashcima.scene.semantic.Note import Note
from smashcima.scene.semantic.BeamedGroup import BeamedGroup
from smashcima.scene.semantic.BeamValue import BeamValue
from smashcima.scene.semantic.ScoreMeasure import ScoreMeasure
from smashcima.scene.semantic.Part import Part
from smashcima.scene.semantic.Chord import Chord
from smashcima.scene.semantic.Measure import Measure
from smashcima.scene.semantic.StemValue import StemValue
from smashcima.scene.semantic.TypeDuration import TypeDuration
from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.visual.Notehead import Notehead
from smashcima.scene.visual.NoteheadSide import NoteheadSide
from smashcima.scene.visual.Stem import Stem
from smashcima.scene.visual.Beam import Beam
from smashcima.scene.visual.BeamHook import BeamHook
from smashcima.scene.visual.BeamCoordinateSystem import BeamCoordinateSystem
from smashcima.geometry.Vector2 import Vector2
from smashcima.geometry.Point import Point
from smashcima.synthesis.LineSynthesizer import LineSynthesizer
from smashcima.scene.SmuflLabels import SmuflLabels
from typing import List, Optional
from collections import Counter
import cv2
import numpy as np
import random
import math


MAX_BEAM_SLOPE = 0.35
BEAM_SLOPE_JITTER = 0.1
BEAM_SPACING = 1.5 # TODO: guess or synth somehow
HOOK_LENGTH = 3 # TODO: guess of synth somehow


def softclamp(x: float, limit: float) -> float:
    """Acts as a symmetric clamp to keep (-limit <= x <= limit), but softly."""
    return math.tanh(x / limit) * limit


class BeamStemSynthesizer:
    def __init__(self, line_synthesizer: LineSynthesizer, rng: random.Random):
        self.line_synthesizer = line_synthesizer
        self.rng = rng

    def synthesize_beams_and_stems_for_measure(
        self,
        page_space: AffineSpace,
        score_measure: ScoreMeasure
    ):
        # get all chords and beamed groups
        chords: List[Chord] = []
        groups: List[BeamedGroup] = []

        for score_event in score_measure.events:
            for event in score_event.events:
                for durable in event.durables:

                    if not isinstance(durable, Note):
                        continue

                    chord = Chord.of_note(durable)
                    if chord not in chords:
                        chords.append(chord)
                    
                    group = BeamedGroup.of_chord_or_none(chord)
                    if group is not None and group not in groups:
                        groups.append(group)
        
        # synthesize stems for all chords
        for chord in chords:
            self.synthesize_stem(page_space, chord)

        # synthesize beam for each beamed group
        for group in groups:
            self.synthesize_beams(page_space, group)

        # adjust stems for beamed groups
        for group in groups:
            self.adjust_stems_for_beam_group(page_space, group)

    def synthesize_stem(self, page_space: AffineSpace, chord: Chord):
        # notehead in the middle of the stem (the "top" notehead)
        center_notehead = Notehead.of_note(
            chord.notes[-1 if chord.stem_value == StemValue.up else 0]
        )

        # notehead at the base of the stem (the "bottom" notehead)
        base_notehead = Notehead.of_note(
            chord.notes[0 if chord.stem_value == StemValue.up else -1]
        )

        # get stem orientation
        stem_value = chord.stem_value or self.infer_stem_orientation(chord)
        assert stem_value in [StemValue.none, StemValue.up, StemValue.down], \
            f"Invalid stem value {stem_value}, must be only up or down or none."
        
        # if no stem, do not render one
        if stem_value == StemValue.none:
            return

        # get both stem points
        stem_base = page_space.transform_from(base_notehead.glyph.space).apply_to(
            self.get_stem_base_point(base_notehead, stem_value)
        )
        stem_center = page_space.transform_from(center_notehead.glyph.space).apply_to(
            self.get_stem_base_point(center_notehead, stem_value)
        )
        stem_tip = self.get_stem_tip_point(stem_center, stem_value)

        # synthesize the stem glyph
        glyph = self.line_synthesizer.synthesize_line(
            label=SmuflLabels.stem.value,
            parent_space=page_space,
            start_point=stem_base,
            end_point=stem_tip
        )
        Stem(
            glyph=glyph,
            chord=chord
        )
    
    def infer_stem_orientation(self, chord: Chord) -> StemValue:
        # whole notes have no stem
        type_duration = chord.get_type_duration()
        if type_duration == TypeDuration.whole:
            return StemValue.none
        
        # for all other notes, the inference is complicated:
        # 1) pitch position
        # 2) average pitch position per beamed group?
        # 3) two voices too close

        # NOTE: could be implemented in the future,
        # or can be subclassed and overriden by the user
        event = chord.get_event()
        onset = event.fractional_measure_onset
        measure = Measure.of_event(event)
        part = Part.of_measure(measure)
        measure_index = part.measures.index(measure)
        raise NotImplementedError(
            "Stem orientation inference is not supported. " + \
            "Provide explicit stem orientation in the scene semantic data. " + \
            f"Measure index: {measure_index}, onset: {onset}"
        )

    def get_stem_base_point(
        self,
        notehead: Notehead,
        orientation: StemValue
    ) -> Point:
        """Returns the point where the stem begins, in the notehead's space"""
        assert orientation in [StemValue.up, StemValue.down], \
            "Only up/down stem values are expected when synthesizing stems."

        side = notehead.up_stem_attachment_side \
            if orientation == StemValue.up \
            else notehead.down_stem_attachment_side
        
        if side is None:
            raise Exception(
                f"There should not be a '{orientation}' stem for the notehead"
            )

        # TODO: pull from some distribution
        if side == NoteheadSide.left:
            return Point(
                x=-notehead.glyph.get_bbox_in_space(notehead.glyph.space).width / 2,
                y=0
            )
        elif side == NoteheadSide.right:
            return Point(
                x=notehead.glyph.get_bbox_in_space(notehead.glyph.space).width / 2,
                y=0
            )
        
        raise Exception(f"Unexpected notehead side value: {side}")
    
    def get_stem_tip_point(
        self,
        base_point: Point,
        orientation: StemValue
    ) -> Point:
        """Returns the stem tip point in paper space, given the base point
        also in paper space"""
        # TODO: sample from a learned distribution
        delta_vector = Vector2(0.0, -5.0)

        if orientation == StemValue.down:
            delta_vector = -delta_vector
        
        return Point.from_origin_vector(base_point.vector + delta_vector)

    def synthesize_beams(self, page_space: AffineSpace, group: BeamedGroup):
        assert len(group.chords) >= 2, \
            "There must be at least 2 chords in a beamed group"

        stems = [Stem.of_chord(ch) for ch in group.chords]
        tips = [s.tip.transform_to(page_space) for s in stems]
        up_stem_count = len(
            [s for s in stems if s.chord.stem_value == StemValue.up]
        )
        down_stem_count = len(
            [s for s in stems if s.chord.stem_value == StemValue.down]
        )
        assert up_stem_count + down_stem_count == len(stems), \
            "Beamed group must contain only up/down stems"
        
        # === 1) compute the slope of the beam ===

        # get the slope of the beam by fitting a line through the points
        slope = 0.0
        line = cv2.fitLine(
            np.array([list(tip) for tip in tips]),
            cv2.DIST_L2, 0, 0, 0
        ).flatten()
        slope = line[1] / line[0]
        
        # randomize slope if too straight
        if abs(slope) < 0.01:
            slope = self.rng.uniform(-BEAM_SLOPE_JITTER, BEAM_SLOPE_JITTER)
        
        # soft-clamp the slope to a reasonable value
        slope = softclamp(slope, MAX_BEAM_SLOPE)

        # === 2) compute beam vertical placement ===

        # If all stems pointed up, we place the beam from above so high,
        # that it does not squish a single stem. It only lengthens them.
        # Same if they all point down. If they have mixed directions,
        # we do both of these and then take the weighted average of the
        # two possible beam placements.

        def get_q(sub_tips: List[Point], stem_up: bool) -> Optional[float]:
            """Returns the "q" term of "kx+q" line for a list of tip points"""
            if len(sub_tips) == 0:
                return None
            
            # define the reference line
            origin = tips[0]
            offset = 0.0
            def _f(x: float) -> float:
                "The function of the beam line"
                return origin.y + (x - origin.x) * slope + offset
            
            if stem_up:
                offset -= max(
                    _f(sub_tip.x) - sub_tip.y
                    for sub_tip in sub_tips
                )
            else:
                offset += max(
                    sub_tip.y - _f(sub_tip.x)
                    for sub_tip in sub_tips
                )
            
            return _f(0)
        
        # get the two "q" constnat offsets of the line at x=0
        # based on stems pointing up or down
        q_for_up = get_q([
            tips[i] for i in range(len(stems))
            if stems[i].chord.stem_value == StemValue.up
        ], stem_up=True)
        q_for_down = get_q([
            tips[i] for i in range(len(stems))
            if stems[i].chord.stem_value == StemValue.down
        ], stem_up=False)

        # if there are both movements, do a weighted average,
        # else take the computed movement
        if q_for_up is not None and q_for_down is not None:
            q = (
                (q_for_up * up_stem_count + q_for_down * down_stem_count)
                / len(stems)
            )
        elif q_for_up is not None:
            q = q_for_up
        elif q_for_down is not None:
            q = q_for_down
        else:
            raise Exception("This should never be raised.")
        
        f = BeamCoordinateSystem(
            beamed_group=group,
            page_space=page_space,
            k=slope,
            q=q,
            beam_spacing=BEAM_SPACING
        )

        # === 3) synthesize beam glyphs ===

        def _determine_beam_orientation(chords: List[Chord]) -> StemValue:
            key, count = Counter(
                [ch.stem_value for ch in chords]
            ).most_common(1)[0]
            assert key is not None, "There must be at least one stem"
            return key

        # beams
        for beam_number, chords in group.iterate_beams():
            start_tip = tips[group.chords.index(chords[0])]
            end_tip = tips[group.chords.index(chords[-1])]
            stem_value = _determine_beam_orientation(chords)
            glyph = self.line_synthesizer.synthesize_line(
                label=SmashcimaLabels.beam.value,
                parent_space=page_space,
                start_point=f.point(start_tip.x, beam_number, stem_value),
                end_point=f.point(end_tip.x, beam_number, stem_value)
            )
            Beam(
                glyph=glyph,
                beamed_group=group,
                chords=chords,
                beam_number=beam_number
            )
        
        # hooks
        for beam_number, chord, hook_type in group.iterate_hooks():
            start_x = tips[group.chords.index(chord)].x
            end_x = start_x + (
                HOOK_LENGTH
                if hook_type == BeamValue.forward_hook else
                -HOOK_LENGTH
            )
            glyph = self.line_synthesizer.synthesize_line(
                label=SmashcimaLabels.beamHook.value,
                parent_space=page_space,
                start_point=f.point(start_x, beam_number, stem_value),
                end_point=f.point(end_x, beam_number, stem_value)
            )
            BeamHook(
                glyph=glyph,
                beamed_group=group,
                chord=chord,
                beam_number=beam_number,
                hook_type=hook_type
            )

    def adjust_stems_for_beam_group(
        self,
        page_space: AffineSpace,
        group: BeamedGroup
    ):
        stems = [Stem.of_chord(ch) for ch in group.chords]
        
        f = BeamCoordinateSystem.of_beamed_group(group)
        
        for old_stem in stems:
            # get the new placement of the stem
            start_point = old_stem.base.transform_to(page_space)
            end_point = f.point(old_stem.tip.transform_to(page_space).x)

            # synthesize new sprites and points
            glyph = self.line_synthesizer.synthesize_line(
                label=old_stem.glyph.label,
                parent_space=page_space,
                start_point=start_point,
                end_point=end_point
            )
            new_stem = Stem(
                glyph=glyph,
                chord=old_stem.chord
            )

            # remove the old stem from scene
            old_stem.detach()


from dataclasses import dataclass
from .Event import Event
from .Measure import Measure
from typing import List, Dict
from fractions import Fraction


@dataclass
class ScoreEvent:
    """An event that spans all score parts; not in the scene, just a view"""

    fractional_measure_onset: Fraction
    """How many quarter notes from the beginning of the measure this event occus"""

    events: List[Event]
    """All measure-events that have the same onset. The number of these events
    may be less than the number of parts (not all measure-events are present
    in all the parts). The order is thus also not defined."""

    @staticmethod
    def merge_from_measures(measures: List[Measure]) -> List["ScoreEvent"]:
        dictionary: Dict[Fraction, ScoreEvent] = {}
        
        for measure in measures:
            for event in measure.events:
                key = event.fractional_measure_onset
                if key not in dictionary:
                    dictionary[key] = ScoreEvent(
                        fractional_measure_onset=key,
                        events=[]
                    )
                dictionary[key].events.append(event)

        values = list(dictionary.values())
        values.sort(key=lambda e: e.fractional_measure_onset)
        return values



from typing import Any, Callable, Tuple, Type, TypeVar, List, Optional
from dataclasses import dataclass, field

from smashcima.scene.nameof_via_dummy import nameof_via_dummy


T = TypeVar("T", bound="SceneObject")


class SceneRelationshipResolutionException(Exception):
    """Thrown by the `.of`, `.many_of`, and `.of_or_none` query methods"""
    pass


class Link:
    """Describes a named, oriented link between two scene objects"""
    source: "SceneObject"
    target: "SceneObject"
    name: str

    def __init__(self, source: "SceneObject", target: "SceneObject", name: str):
        self.source = source
        self.target = target
        self.name = name
    
    def __eq__(self, other) -> bool:
        if not isinstance(other, Link):
            return False
        if self.source is not other.source:
            return False
        if self.target is not other.target:
            return False
        if self.name != other.name:
            return False
        return True
    
    def __repr__(self) -> str:
        return f"{self.source.__class__.__name__}-->" + \
            f"{self.target.__class__.__name__}"

    def attach(self):
        """Add the link into the graph"""
        self.source.outlinks.append(self)
        self.target.inlinks.append(self)

    def detach(self):
        """Remove the link from the graph"""
        self.source.outlinks.remove(self)
        self.target.inlinks.remove(self)


@dataclass
class SceneObject:
    inlinks: List[Link] = field(default_factory=list, init=False, repr=False)
    outlinks: List[Link] = field(default_factory=list, init=False, repr=False)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "inlinks":
            pass
        elif name == "outlinks":
            pass
        elif isinstance(value, SceneObject):
            self._destroy_outlinks_for(name)
            Link(source=self, target=value, name=name).attach()
        elif isinstance(value, list) or isinstance(value, set):
            # TODO: hook into mutation methods or freeze the instances!
            self._destroy_outlinks_for(name)
            for item in value:
                if isinstance(item, SceneObject):
                    Link(source=self, target=item, name=name).attach()
        else:
            self._destroy_outlinks_for(name)

        super().__setattr__(name, value)
    
    def _destroy_outlinks_for(self, name: str):
        for link in list(self.outlinks):
            if link.name == name:
                link.detach()

    ########################
    # Relationship queries #
    ########################

    @classmethod
    def _of_impl(
        cls: Type[T],
        subject: "SceneObject",
        name_probe: Callable[[T], Any]
    ) -> Tuple[List[T], str]:
        name = nameof_via_dummy(cls, name_probe)
        return (
            [
                link.source for link in subject.inlinks
                if isinstance(link.source, cls)
                    and (name is None or link.name == name)
            ],
            name
        )

    @classmethod
    def of(
        cls: Type[T], 
        subject: "SceneObject",
        name_probe: Callable[[T], Any]
    ) -> T:
        if subject is None:
            raise ValueError("Passed in subject cannot be None")
        
        candidates, name = cls._of_impl(subject, name_probe)
        
        if len(candidates) > 1:
            raise SceneRelationshipResolutionException(
                f"There are more than one {cls} linking to " + \
                f"{subject} via name {name}."
            )
        
        if len(candidates) == 0:
            raise SceneRelationshipResolutionException(
                f"There are no {cls} linking to " + \
                f"{subject} via name {name}."
            )
        
        return candidates[0]

    @classmethod
    def of_or_none(
        cls: Type[T],
        subject: Optional["SceneObject"],
        name_probe: Callable[[T], Any]
    ) -> Optional[T]:
        if subject is None:
            return None
        
        candidates, name = cls._of_impl(subject, name_probe)
        
        if len(candidates) > 1:
            raise SceneRelationshipResolutionException(
                f"There are more than one {cls} linking to " + \
                f"{subject} via name {name}."
            )
        
        if len(candidates) == 0:
            return None
        
        return candidates[0]

    @classmethod
    def many_of(
        cls: Type[T],
        subject: "SceneObject",
        name_probe: Callable[[T], Any]
    ) -> List[T]:
        if subject is None:
            raise ValueError("Passed in subject cannot be None")
        
        candidates, _ = cls._of_impl(subject, name_probe)
        
        return candidates



import abc


class Column(abc.ABC):
    """Public interface of a layout column"""
    
    # NOTE: The construction of the glyph is not codified, because each one
    # might be constructed in a different manner. What matters here is the
    # methods used by the layout synthesizer to position and space out
    # all the columns.

    def __init__(self):
        self.time_position = 0
        "Where on the staff (staves) is this column placed"
        
        self.width = 0
        "Total width in millimeters of the column"
        
        self.left_width = 0
        "Width left of the origin vertical line"

        self.right_width = 0
        "Width right of the origin vertical line"

        self.flex_grow = 1
        "How much does this column grow when stretched in relation to others"

        self.flex_shrink = 1
        "How much does this column get shrunk when the system is overflowing"

    @abc.abstractmethod
    def position_glyphs(self):
        """Called to place glyphs onto staves based on the current time_position"""
        raise NotImplementedError

    @abc.abstractmethod
    def recalculate_dimensions(self):
        """Should be called at the end of glyph positioning to update width values"""
        raise NotImplementedError

    @abc.abstractmethod
    def detach(self):
        """Removes the column from the scene"""
        raise NotImplementedError



import abc
from typing import List
from .Column import Column
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.Glyph import Glyph
from smashcima.scene.Sprite import Sprite
from smashcima.geometry.Vector2 import Vector2
from smashcima.geometry.Rectangle import Rectangle
import random


class ColumnBase(Column, metaclass=abc.ABCMeta):
    """Base class for layout columns that implements common logic,
    such as dimension calculation from glyph list"""
    def __init__(self, staves: List[StaffVisual], rng_seed: float):
        super().__init__()

        self.staves = staves
        "Reference to the list of stafflines we render onto"

        self.glyphs: List[Glyph] = []
        "Glyphs tracked in this column"

        self.rng_seed = rng_seed
        "Seed for the local RNG used for glyph positioning"

        self.rng = random.Random(self.rng_seed)
        "The RNG that should be used for randomizing glyph positioning"

        self.__post_init__()
    
    def __post_init__(self):
        # can be overriden in child classes
        pass
    
    def add_glyph(self, glyph: Glyph):
        """Adds a glyph into the column, the glyph must be in some staff space"""
        # check that the glyph is properly attached to stafflines
        self.get_staff_of_glyph(glyph)
        if glyph not in self.glyphs:
            self.glyphs.append(glyph)
    
    def get_staff_of_glyph(self, glyph: Glyph) -> StaffVisual:
        """Resolves the stafflines object for a given glyph"""
        for staff in self.staves:
            if staff.space is glyph.space.parent_space:
                return staff
        raise Exception(
            "Given glyph is not attached to any of the stafflines objects"
        )

    def position_glyphs(self):
        """Positions glyphs on staves according to the column's time position"""
        # reset RNG before any positioning
        self.rng = random.Random(self.rng_seed)
        self._position_glyphs()
        self.recalculate_dimensions()

    @abc.abstractmethod
    def _position_glyphs(self):
        """Glyph positioning implementation"""
        raise NotImplementedError

    def recalculate_dimensions(self):
        # reset all
        self.width = 0
        self.left_width = 0
        self.right_width = 0

        # map sprite corners to the stafflines space
        # relative to its time position
        points: List[Vector2] = []

        # for all staves
        for staff in self.staves:
            column_origin = staff.staff_coordinate_system.get_transform(
                0,
                self.time_position
            ).apply_to(Vector2(0, 0))

            # for all glyphs in that staff
            for glyph in self.glyphs:
                if glyph.space.parent_space is not staff.space:
                    continue

                # get all sprites in that glyph and transform their corners
                # into the staff space
                for (sprite, transform) in Sprite.traverse_sprites(
                    glyph.space,
                    include_pixels_transform=True,
                    include_sprite_transform=True,
                    include_root_space_transform=True
                ):
                    corners = [
                        transform.apply_to(Vector2(0, 0)),
                        transform.apply_to(Vector2(sprite.pixel_width, 0)),
                        transform.apply_to(Vector2(0, sprite.pixel_height)),
                        transform.apply_to(
                            Vector2(sprite.pixel_width, sprite.pixel_height)
                        )
                    ]
                    corners = [p - column_origin for p in corners]
                    points += corners

        # recalculate widths
        if len(points) > 0:
            self.left_width = -min(p.x for p in points)
            self.right_width = max(p.x for p in points)
            self.width = self.left_width + self.right_width
    
    def detach(self):
        for glyph in self.glyphs:
            glyph.detach()

    def place_debug_boxes(self):
        """For debugging purposes - places sprites that visualize the column"""
        for staff in self.staves:
            topleft = staff.staff_coordinate_system.get_transform(
                pitch_position=4,
                time_position=self.time_position - self.left_width
            ).apply_to(Vector2(0, 0))
            bottomright = staff.staff_coordinate_system.get_transform(
                pitch_position=-4,
                time_position=self.time_position + self.right_width
            ).apply_to(Vector2(0, 0))
            Sprite.debug_box(
                space=staff.space,
                rectangle=Rectangle(
                    x=topleft.x,
                    y=topleft.y,
                    width=bottomright.x - topleft.x,
                    height=bottomright.y - topleft.y
                ),
                fill_color=(0, 0, 0, 0),
                border_width=0.2 # mm
            )



from dataclasses import dataclass
from .ClefSign import ClefSign
from .Pitch import Pitch


@dataclass
class Clef:
    """Represents a music notation clef"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/clef/

    sign: ClefSign
    "What type of clef is this"

    line: int
    "Position of the clef, 1 is bottom staffline, numbering goes up"

    after_barline: bool = False
    """Should the clef be rendered after the barline? If not (default),
    the clef is rendered at the end of previous measure (if its onset is 0).
    Ignored for mid-measure clefs."""

    @staticmethod
    def clef_line_to_pitch_position(line: int) -> int:
        """Given the staff line number, return the pitch position of that
        staff line. Pitch position is 0 on the center line, increases up and
        1 step is one half-space, i.e. the top line is 4 and the bottom line
        is -4."""
        # NOTE: we assume 5 line staff
        return (line - 3) * 2
    
    @staticmethod
    def origin_pitch(sign: ClefSign) -> Pitch:
        """Which pitch sits at the origin of the clef (e.g. G4 for G clef)"""
        if sign == ClefSign.G:
            return Pitch.parse("4", "G")
        if sign == ClefSign.F:
            return Pitch.parse("3", "F")
        if sign == ClefSign.C:
            return Pitch.parse("4", "C")
        raise Exception(f"Unknown clef sign {sign}")
    
    def pitch_to_pitch_position(self, pitch: Pitch) -> int:
        """Returns the stafflines-pitch-position for the given note pitch
        if this clef is currently active for the staff"""
        origin_pitch_position = Clef.clef_line_to_pitch_position(self.line)
        origin_linear_pitch = Clef.origin_pitch(self.sign) \
            .get_linear_pitch(with_alters=False)
        given_linear_pitch = pitch.get_linear_pitch(with_alters=False)
        return origin_pitch_position \
            + (given_linear_pitch - origin_linear_pitch)

    def pitch_position_to_pitch(self, pitch_position: int) -> Pitch:
        """Returns the note pitch on a given stafflines-pitch-position
        if this clef is currently active for the staff"""
        origin_pitch_position = Clef.clef_line_to_pitch_position(self.line)
        origin_linear_pitch = Clef.origin_pitch(self.sign) \
            .get_linear_pitch(with_alters=False)
        linear_pitch = (pitch_position - origin_pitch_position) \
            + origin_linear_pitch
        return Pitch.from_linear_pitch(linear_pitch)



from .Vector2 import Vector2


class Point:
    """Geometric 2D point"""
    def __init__(self, x: float, y: float):
        self.vector = Vector2(float(x), float(y))
        "The vector from origin to the point position that defines this point"
    
    @staticmethod
    def from_origin_vector(vector: Vector2) -> "Point":
        return Point(vector.x, vector.y)

    @property
    def x(self) -> float:
        return self.vector.x

    @property
    def y(self) -> float:
        return self.vector.y

    @property
    def left(self) -> float:
        return self.vector.x

    @property
    def top(self) -> float:
        return self.vector.y
    
    def __repr__(self):
        return f"Point({self.x}, {self.y})"
    
    def __neg__(self):
        return Point(-self.x, -self.y)
    
    def __iter__(self):
        yield self.x
        yield self.y



from typing import Union


MILLIMETERS_IN_INCH = 24.4
"The number of millimeters in one inch"


INCHES_IN_MILLIMETER = 1 / MILLIMETERS_IN_INCH
"The number of inches in one millimeter"


def px_to_mm(pixels: Union[float, int], dpi: Union[float, int]) -> float:
    """Converts pixels to millimeters given a DPI"""
    return (float(pixels) / float(dpi)) * MILLIMETERS_IN_INCH


def mm_to_px(millimeters: Union[float, int], dpi: Union[float, int]) -> float:
    """Converts millimeters to pixels given a DPI"""
    return float(millimeters) * INCHES_IN_MILLIMETER * float(dpi)



from pathlib import Path
from muscima.io import parse_cropobject_list, CropObject
from typing import List, Dict
from .PointCloud import PointCloud
import re


class MppPage:
    """
    The CVC-MUSCIMA dataset consists of 1000 scanned pages.
    This class represents one of them, annotated according to the MUSCIMA++
    standard. (Note: MUSCIMA++ contains only 140 pages out of the 1000)
    """
    def __init__(
        self,
        crop_objects: List[CropObject],
        mpp_writer: int,
        mpp_piece: int
    ):
        self.crop_objects = crop_objects
        "List of cropobjects as they are loaded by the muscima library"

        self.id_lookup: Dict[int, CropObject] = {
            c.objid: c
            for c in self.crop_objects
        }
        "Dictionary for fast id lookup"

        self.point_cloud = PointCloud()
        "Stores extracted points in crop objects"

        assert mpp_writer >= 1 and mpp_writer <= 50
        self.mpp_writer = mpp_writer
        "Writer index (1 to 50) from the MUSCIMA++ dataset"

        assert mpp_piece >= 1 and mpp_piece <= 20
        self.mpp_piece = mpp_piece
        "Music piece index (1 to 20) from the MUSCIMA++ dataset"

    @staticmethod
    def load(path: Path) -> "MppPage":
        crop_objects: List[CropObject] = parse_cropobject_list(path)

        m = re.match(r"CVC-MUSCIMA_W-(\d+)_N-(\d+)_D-ideal", path.stem)
        mpp_writer = int(m.group(1))
        mpp_piece = int(m.group(2))

        return MppPage(crop_objects, mpp_writer, mpp_piece)
    
    def get(self, objid: int) -> CropObject:
        """Looks up a crop object by its integer ID"""
        return self.id_lookup[objid]
    
    def get_outlink_to(self, obj: CropObject, clsname: str) -> CropObject:
        """
        Returns the crop object at the end of any outlink
        with the given classname
        """
        for l in obj.outlinks:
            resolved_link = self.get(l)
            if resolved_link.clsname == clsname:
                return resolved_link
        raise Exception("Object has no outlink of requested clsname")
    
    def get_inlink_from(self, obj: CropObject, clsname: str) -> CropObject:
        """
        Returns the crop object at the end of any inlink
        with the given classname
        """
        for l in obj.inlinks:
            resolved_link = self.get(l)
            if resolved_link.clsname == clsname:
                return resolved_link
        raise Exception("Object has no inlinks of requested clsname")

    def has_outlink_to(self, obj: CropObject, clsname: str) -> bool:
        """
        Tests that there is a crop object at the end of any outlink
        with the given classname
        """
        for l in obj.outlinks:
            resolved_link = self.get(l)
            if resolved_link.clsname == clsname:
                return True
        return False

    def has_inlink_from(self, obj: CropObject, clsname: str) -> bool:
        """
        Tests that there is a crop object at the end of any inlink
        with the given classname
        """
        for l in obj.inlinks:
            resolved_link = self.get(l)
            if resolved_link.clsname == clsname:
                return True
        return False
